@@ -14,12 +14,25 @@
 
 import contextlib
 import os
+import platform
 import socket
+import sys
 import threading
 
+try:
+    import ssl
+    from ssl import SSLError
+    _HAVE_SNI = getattr(ssl, 'HAS_SNI', False)
+except ImportError:
+    _HAVE_SNI = False
+    class SSLError(socket.error):
+        pass
+
+
 from bson import DEFAULT_CODEC_OPTIONS
-from bson.py3compat import itervalues
-from pymongo import auth, helpers, thread_util
+from bson.py3compat import imap, itervalues, _unicode, integer_types
+from bson.son import SON
+from pymongo import auth, helpers, thread_util, __version__
 from pymongo.common import MAX_MESSAGE_SIZE
 from pymongo.errors import (AutoReconnect,
                             ConnectionFailure,
@@ -32,10 +45,193 @@ from pymongo.ismaster import IsMaster
 from pymongo.monotonic import time as _time
 from pymongo.network import (command,
                              receive_message,
-                             socket_closed)
+                             SocketChecker)
 from pymongo.read_concern import DEFAULT_READ_CONCERN
 from pymongo.read_preferences import ReadPreference
 from pymongo.server_type import SERVER_TYPE
+# Always use our backport so we always have support for IP address matching
+from pymongo.ssl_match_hostname import match_hostname, CertificateError
+
+# For SNI support. According to RFC6066, section 3, IPv4 and IPv6 literals are
+# not permitted for SNI hostname.
+try:
+    from ipaddress import ip_address
+    def is_ip_address(address):
+        try:
+            ip_address(_unicode(address))
+            return True
+        except (ValueError, UnicodeError):
+            return False
+except ImportError:
+    if hasattr(socket, 'inet_pton') and socket.has_ipv6:
+        # Most *nix, recent Windows
+        def is_ip_address(address):
+            try:
+                # inet_pton rejects IPv4 literals with leading zeros
+                # (e.g. 192.168.0.01), inet_aton does not, and we
+                # can connect to them without issue. Use inet_aton.
+                socket.inet_aton(address)
+                return True
+            except socket.error:
+                try:
+                    socket.inet_pton(socket.AF_INET6, address)
+                    return True
+                except socket.error:
+                    return False
+    else:
+        # No inet_pton
+        def is_ip_address(address):
+            try:
+                socket.inet_aton(address)
+                return True
+            except socket.error:
+                if ':' in address:
+                    # ':' is not a valid character for a hostname. If we get
+                    # here a few things have to be true:
+                    #   - We're on a recent version of python 2.7 (2.7.9+).
+                    #     2.6 and older 2.7 versions don't support SNI.
+                    #   - We're on Windows XP or some unusual Unix that doesn't
+                    #     have inet_pton.
+                    #   - The application is using IPv6 literals with TLS, which
+                    #     is pretty unusual.
+                    return True
+                return False
+
+try:
+    from fcntl import fcntl, F_GETFD, F_SETFD, FD_CLOEXEC
+    def _set_non_inheritable_non_atomic(fd):
+        """Set the close-on-exec flag on the given file descriptor."""
+        flags = fcntl(fd, F_GETFD)
+        fcntl(fd, F_SETFD, flags | FD_CLOEXEC)
+except ImportError:
+    # Windows, various platforms we don't claim to support
+    # (Jython, IronPython, ...), systems that don't provide
+    # everything we need from fcntl, etc.
+    def _set_non_inheritable_non_atomic(dummy):
+        """Dummy function for platforms that don't provide fcntl."""
+        pass
+
+_MAX_TCP_KEEPIDLE = 300
+_MAX_TCP_KEEPINTVL = 10
+_MAX_TCP_KEEPCNT = 9
+
+if sys.platform == 'win32':
+    try:
+        import _winreg as winreg
+    except ImportError:
+        import winreg
+
+    try:
+        with winreg.OpenKey(
+                winreg.HKEY_LOCAL_MACHINE,
+                r"SYSTEM\CurrentControlSet\Services\Tcpip\Parameters") as key:
+            _DEFAULT_TCP_IDLE_MS, _ = winreg.QueryValueEx(key, "KeepAliveTime")
+            _DEFAULT_TCP_INTERVAL_MS, _ = winreg.QueryValueEx(
+                key, "KeepAliveInterval")
+            # Make sure these are integers.
+            if not isinstance(_DEFAULT_TCP_IDLE_MS, integer_types):
+                raise ValueError
+            if not isinstance(_DEFAULT_TCP_INTERVAL_MS, integer_types):
+                raise ValueError
+    except (OSError, ValueError):
+        # We could not check the default values so do not attempt to override.
+        def _set_keepalive_times(dummy):
+            pass
+    else:
+        def _set_keepalive_times(sock):
+            idle_ms = min(_DEFAULT_TCP_IDLE_MS, _MAX_TCP_KEEPIDLE * 1000)
+            interval_ms = min(_DEFAULT_TCP_INTERVAL_MS,
+                              _MAX_TCP_KEEPINTVL * 1000)
+            if (idle_ms < _DEFAULT_TCP_IDLE_MS or
+                    interval_ms < _DEFAULT_TCP_INTERVAL_MS):
+                sock.ioctl(socket.SIO_KEEPALIVE_VALS,
+                           (1, idle_ms, interval_ms))
+else:
+    def _set_tcp_option(sock, tcp_option, max_value):
+        if hasattr(socket, tcp_option):
+            sockopt = getattr(socket, tcp_option)
+            default = sock.getsockopt(socket.SOL_TCP, sockopt)
+            if default > max_value:
+                sock.setsockopt(socket.SOL_TCP, sockopt, max_value)
+
+    def _set_keepalive_times(sock):
+        _set_tcp_option(sock, 'TCP_KEEPIDLE', _MAX_TCP_KEEPIDLE)
+        _set_tcp_option(sock, 'TCP_KEEPINTVL', _MAX_TCP_KEEPINTVL)
+        _set_tcp_option(sock, 'TCP_KEEPCNT', _MAX_TCP_KEEPCNT)
+
+_METADATA = SON([
+    ('driver', SON([('name', 'PyMongo'), ('version', __version__)])),
+])
+
+if sys.platform.startswith('linux'):
+    # platform.linux_distribution was deprecated in Python 3.5.
+    if sys.version_info[:2] < (3, 5):
+        # Distro name and version (e.g. Ubuntu 16.04 xenial)
+        _name = ' '.join([part for part in
+                          platform.linux_distribution() if part])
+    else:
+        _name = platform.system()
+    _METADATA['os'] = SON([
+        ('type', platform.system()),
+        ('name', _name),
+        ('architecture', platform.machine()),
+        # Kernel version (e.g. 4.4.0-17-generic).
+        ('version', platform.release())
+    ])
+elif sys.platform == 'darwin':
+    _METADATA['os'] = SON([
+        ('type', platform.system()),
+        ('name', platform.system()),
+        ('architecture', platform.machine()),
+        # (mac|i|tv)OS(X) version (e.g. 10.11.6) instead of darwin
+        # kernel version.
+        ('version', platform.mac_ver()[0])
+    ])
+elif sys.platform == 'win32':
+    _METADATA['os'] = SON([
+        ('type', platform.system()),
+        # "Windows XP", "Windows 7", "Windows 10", etc.
+        ('name', ' '.join((platform.system(), platform.release()))),
+        ('architecture', platform.machine()),
+        # Windows patch level (e.g. 5.1.2600-SP3)
+        ('version', '-'.join(platform.win32_ver()[1:3]))
+    ])
+elif sys.platform.startswith('java'):
+    _name, _ver, _arch = platform.java_ver()[-1]
+    _METADATA['os'] = SON([
+        # Linux, Windows 7, Mac OS X, etc.
+        ('type', _name),
+        ('name', _name),
+        # x86, x86_64, AMD64, etc.
+        ('architecture', _arch),
+        # Linux kernel version, OSX version, etc.
+        ('version', _ver)
+    ])
+else:
+    # Get potential alias (e.g. SunOS 5.11 becomes Solaris 2.11)
+    _aliased = platform.system_alias(
+        platform.system(), platform.release(), platform.version())
+    _METADATA['os'] = SON([
+        ('type', platform.system()),
+        ('name', ' '.join([part for part in _aliased[:2] if part])),
+        ('architecture', platform.machine()),
+        ('version', _aliased[2])
+    ])
+
+if platform.python_implementation().startswith('PyPy'):
+    _METADATA['platform'] = ' '.join(
+        (platform.python_implementation(),
+         '.'.join(imap(str, sys.pypy_version_info)),
+         '(Python %s)' % '.'.join(imap(str, sys.version_info))))
+elif sys.platform.startswith('java'):
+    _METADATA['platform'] = ' '.join(
+        (platform.python_implementation(),
+         '.'.join(imap(str, sys.version_info)),
+         '(%s)' % ' '.join((platform.system(), platform.release()))))
+else:
+    _METADATA['platform'] = ' '.join(
+        (platform.python_implementation(),
+         '.'.join(imap(str, sys.version_info))))
 
 
 # If the first getaddrinfo call of this interpreter's life is on a thread,
@@ -44,12 +240,6 @@ from pymongo.server_type import SERVER_TYPE
 # main thread, to avoid the deadlock. See PYTHON-607.
 u'foo'.encode('idna')
 
-try:
-    from ssl import match_hostname, CertificateError
-except ImportError:
-    # These don't require the ssl module
-    from pymongo.ssl_match_hostname import match_hostname, CertificateError
-
 
 def _raise_connection_failure(address, error):
     """Convert a socket.error to ConnectionFailure and raise it."""
@@ -61,6 +251,14 @@ def _raise_connection_failure(address, error):
         msg = '%s: %s' % (host, error)
     if isinstance(error, socket.timeout):
         raise NetworkTimeout(msg)
+    elif isinstance(error, SSLError) and 'timed out' in str(error):
+        # CPython 2.6, 2.7, PyPy 2.x, and PyPy3 do not distinguish network
+        # timeouts from other SSLErrors (https://bugs.python.org/issue10272).
+        # Luckily, we can work around this limitation because the phrase
+        # 'timed out' appears in all the timeout related SSLErrors raised
+        # on the above platforms. CPython >= 3.2 and PyPy3.3 correctly raise
+        # socket.timeout.
+        raise NetworkTimeout(msg)
     else:
         raise AutoReconnect(msg)
 
@@ -71,14 +269,14 @@ class PoolOptions(object):
                  '__connect_timeout', '__socket_timeout',
                  '__wait_queue_timeout', '__wait_queue_multiple',
                  '__ssl_context', '__ssl_match_hostname', '__socket_keepalive',
-                 '__event_listeners')
+                 '__event_listeners', '__appname', '__metadata')
 
     def __init__(self, max_pool_size=100, min_pool_size=0,
                  max_idle_time_ms=None, connect_timeout=None,
                  socket_timeout=None, wait_queue_timeout=None,
                  wait_queue_multiple=None, ssl_context=None,
-                 ssl_match_hostname=True, socket_keepalive=False,
-                 event_listeners=None):
+                 ssl_match_hostname=True, socket_keepalive=True,
+                 event_listeners=None, appname=None):
 
         self.__max_pool_size = max_pool_size
         self.__min_pool_size = min_pool_size
@@ -91,6 +289,10 @@ class PoolOptions(object):
         self.__ssl_match_hostname = ssl_match_hostname
         self.__socket_keepalive = socket_keepalive
         self.__event_listeners = event_listeners
+        self.__appname = appname
+        self.__metadata = _METADATA.copy()
+        if appname:
+            self.__metadata['application'] = {'name': appname}
 
     @property
     def max_pool_size(self):
@@ -173,6 +375,18 @@ class PoolOptions(object):
         """
         return self.__event_listeners
 
+    @property
+    def appname(self):
+        """The application name, for sending with ismaster in server handshake.
+        """
+        return self.__appname
+
+    @property
+    def metadata(self):
+        """A dict of metadata about the application, driver, os, and platform.
+        """
+        return self.__metadata.copy()
+
 
 class SocketInfo(object):
     """Store a socket with some metadata.
@@ -188,7 +402,7 @@ class SocketInfo(object):
         self.address = address
         self.authset = set()
         self.closed = False
-        self.last_checkout = _time()
+        self.last_checkin = _time()
         self.is_writable = ismaster.is_writable if ismaster else None
         self.max_wire_version = ismaster.max_wire_version if ismaster else None
         self.max_bson_size = ismaster.max_bson_size if ismaster else None
@@ -212,7 +426,10 @@ class SocketInfo(object):
                 read_preference=ReadPreference.PRIMARY,
                 codec_options=DEFAULT_CODEC_OPTIONS, check=True,
                 allowable_errors=None, check_keys=False,
-                read_concern=DEFAULT_READ_CONCERN):
+                read_concern=DEFAULT_READ_CONCERN,
+                write_concern=None,
+                parse_write_concern_error=False,
+                collation=None):
         """Execute a command or raise ConnectionFailure or OperationFailure.
 
         :Parameters:
@@ -225,18 +442,33 @@ class SocketInfo(object):
           - `allowable_errors`: errors to ignore if `check` is True
           - `check_keys`: if True, check `spec` for invalid keys
           - `read_concern`: The read concern for this command.
+          - `write_concern`: The write concern for this command.
+          - `parse_write_concern_error`: Whether to parse the
+            ``writeConcernError`` field in the command response.
+          - `collation`: The collation for this command.
         """
         if self.max_wire_version < 4 and not read_concern.ok_for_legacy:
             raise ConfigurationError(
                 'read concern level of %s is not valid '
                 'with a max wire version of %d.'
                 % (read_concern.level, self.max_wire_version))
+        if not (write_concern is None or write_concern.acknowledged or
+                collation is None):
+            raise ConfigurationError(
+                'Collation is unsupported for unacknowledged writes.')
+        if self.max_wire_version >= 5 and write_concern:
+            spec['writeConcern'] = write_concern.document
+        elif self.max_wire_version < 5 and collation is not None:
+            raise ConfigurationError(
+                'Must be connected to MongoDB 3.4+ to use a collation.')
         try:
             return command(self.sock, dbname, spec, slave_ok,
                            self.is_mongos, read_preference, codec_options,
                            check, allowable_errors, self.address,
                            check_keys, self.listeners, self.max_bson_size,
-                           read_concern)
+                           read_concern,
+                           parse_write_concern_error=parse_write_concern_error,
+                           collation=collation)
         except OperationFailure:
             raise
         # Catch socket.error, KeyboardInterrupt, etc. and close ourselves.
@@ -251,7 +483,7 @@ class SocketInfo(object):
         if (self.max_bson_size is not None
                 and max_doc_size > self.max_bson_size):
             raise DocumentTooLarge(
-                "BSON document too large (%d bytes) - the connected server"
+                "BSON document too large (%d bytes) - the connected server "
                 "supports BSON document sizes up to %d bytes." %
                 (max_doc_size, self.max_bson_size))
 
@@ -403,6 +635,8 @@ def _create_connection(address, options):
             raise ConnectionFailure("UNIX-sockets are not supported "
                                     "on this system")
         sock = socket.socket(socket.AF_UNIX)
+        # SOCK_CLOEXEC not supported for Unix sockets.
+        _set_non_inheritable_non_atomic(sock.fileno())
         try:
             sock.connect(host)
             return sock
@@ -420,12 +654,25 @@ def _create_connection(address, options):
     err = None
     for res in socket.getaddrinfo(host, port, family, socket.SOCK_STREAM):
         af, socktype, proto, dummy, sa = res
-        sock = socket.socket(af, socktype, proto)
+        # SOCK_CLOEXEC was new in CPython 3.2, and only available on a limited
+        # number of platforms (newer Linux and *BSD). Starting with CPython 3.4
+        # all file descriptors are created non-inheritable. See PEP 446.
+        try:
+            sock = socket.socket(
+                af, socktype | getattr(socket, 'SOCK_CLOEXEC', 0), proto)
+        except socket.error:
+            # Can SOCK_CLOEXEC be defined even if the kernel doesn't support
+            # it?
+            sock = socket.socket(af, socktype, proto)
+        # Fallback when SOCK_CLOEXEC isn't available.
+        _set_non_inheritable_non_atomic(sock.fileno())
         try:
             sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
             sock.settimeout(options.connect_timeout)
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE,
                             options.socket_keepalive)
+            if options.socket_keepalive:
+                _set_keepalive_times(sock)
             sock.connect(sa)
             return sock
         except socket.error as e:
@@ -453,14 +700,20 @@ def _configured_socket(address, options):
     ssl_context = options.ssl_context
 
     if ssl_context is not None:
+        host = address[0]
         try:
-            sock = ssl_context.wrap_socket(sock)
+            # According to RFC6066, section 3, IPv4 and IPv6 literals are
+            # not permitted for SNI hostname.
+            if _HAVE_SNI and not is_ip_address(host):
+                sock = ssl_context.wrap_socket(sock, server_hostname=host)
+            else:
+                sock = ssl_context.wrap_socket(sock)
         except IOError as exc:
             sock.close()
             raise ConnectionFailure("SSL handshake failed: %s" % (str(exc),))
         if ssl_context.verify_mode and options.ssl_match_hostname:
             try:
-                match_hostname(sock.getpeercert(), hostname=address[0])
+                match_hostname(sock.getpeercert(), hostname=host)
             except CertificateError:
                 sock.close()
                 raise
@@ -504,6 +757,7 @@ class Pool:
 
         self._socket_semaphore = thread_util.create_semaphore(
             self.opts.max_pool_size, max_waiters)
+        self.socket_checker = SocketChecker()
 
     def reset(self):
         with self.lock:
@@ -516,11 +770,11 @@ class Pool:
             sock_info.close()
 
     def remove_stale_sockets(self):
-        with self.lock:
-            if self.opts.max_idle_time_ms is not None:
+        if self.opts.max_idle_time_ms is not None:
+            with self.lock:
                 for sock_info in self.sockets.copy():
-                    age = _time() - sock_info.last_checkout
-                    if age > self.opts.max_idle_time_ms:
+                    idle_time_ms = 1000 * _time() - sock_info.last_checkin
+                    if idle_time_ms > self.opts.max_idle_time_ms:
                         self.sockets.remove(sock_info)
                         sock_info.close()
 
@@ -542,10 +796,18 @@ class Pool:
         try:
             sock = _configured_socket(self.address, self.opts)
             if self.handshake:
-                ismaster = IsMaster(command(sock, 'admin', {'ismaster': 1},
-                                            False, False,
-                                            ReadPreference.PRIMARY,
-                                            DEFAULT_CODEC_OPTIONS))
+                cmd = SON([
+                    ('ismaster', 1),
+                    ('client', self.opts.metadata)
+                ])
+                ismaster = IsMaster(
+                    command(sock,
+                            'admin',
+                            cmd,
+                            False,
+                            False,
+                            ReadPreference.PRIMARY,
+                            DEFAULT_CODEC_OPTIONS))
             else:
                 ismaster = None
             return SocketInfo(sock, self, ismaster, self.address)
@@ -612,28 +874,20 @@ class Pool:
                 # set.pop() isn't atomic in Jython less than 2.7, see
                 # http://bugs.jython.org/issue1854
                 with self.lock:
-                    sock_info, from_pool = self.sockets.pop(), True
+                    # Can raise ConnectionFailure.
+                    sock_info = self.sockets.pop()
             except KeyError:
                 # Can raise ConnectionFailure or CertificateError.
-                sock_info, from_pool = self.connect(), False
-            # If socket is idle, open a new one.
-            if self.opts.max_idle_time_ms is not None:
-                age = _time() - sock_info.last_checkout
-                if age > self.opts.max_idle_time_ms:
-                    sock_info.close()
-                    sock_info, from_pool = self.connect(), False
-
-            if from_pool:
+                sock_info = self.connect()
+            else:
                 # Can raise ConnectionFailure.
                 sock_info = self._check(sock_info)
-
         except:
             self._socket_semaphore.release()
             with self.lock:
                 self.active_sockets -= 1
             raise
 
-        sock_info.last_checkout = _time()
         return sock_info
 
     def return_socket(self, sock_info):
@@ -644,6 +898,7 @@ class Pool:
             if sock_info.pool_id != self.pool_id:
                 sock_info.close()
             elif not sock_info.closed:
+                sock_info.last_checkin = _time()
                 with self.lock:
                     self.sockets.add(sock_info)
 
@@ -652,35 +907,36 @@ class Pool:
             self.active_sockets -= 1
 
     def _check(self, sock_info):
-        """This side-effecty function checks if this pool has been reset since
-        the last time this socket was used, or if the socket has been closed by
-        some external network error, and if so, attempts to create a new socket.
-        If this connection attempt fails we reset the pool and reraise the
+        """This side-effecty function checks if this socket has been idle for
+        for longer than the max idle time, or if the socket has been closed by
+        some external network error, and if so, attempts to create a new
+        socket. If this connection attempt fails we raise the
         ConnectionFailure.
 
         Checking sockets lets us avoid seeing *some*
         :class:`~pymongo.errors.AutoReconnect` exceptions on server
-        hiccups, etc. We only do this if it's been > 1 second since
-        the last socket checkout, to keep performance reasonable - we
-        can't avoid AutoReconnects completely anyway.
+        hiccups, etc. We only check if the socket was closed by an external
+        error if it has been > 1 second since the socket was checked into the
+        pool, to keep performance reasonable - we can't avoid AutoReconnects
+        completely anyway.
         """
-        error = False
-
-        # How long since socket was last checked out.
-        age = _time() - sock_info.last_checkout
-        if (self._check_interval_seconds is not None
-                and (
-                    0 == self._check_interval_seconds
-                    or age > self._check_interval_seconds)):
-            if socket_closed(sock_info.sock):
-                sock_info.close()
-                error = True
-
-        if not error:
-            return sock_info
-        else:
+        # How long since socket was last checked in.
+        idle_time_seconds = _time() - sock_info.last_checkin
+        # If socket is idle, open a new one.
+        if (self.opts.max_idle_time_ms is not None and
+                idle_time_seconds * 1000 > self.opts.max_idle_time_ms):
+            sock_info.close()
             return self.connect()
 
+        if (self._check_interval_seconds is not None and (
+                0 == self._check_interval_seconds or
+                idle_time_seconds > self._check_interval_seconds)):
+            if self.socket_checker.socket_closed(sock_info.sock):
+                sock_info.close()
+                return self.connect()
+
+        return sock_info
+
     def _raise_wait_queue_timeout(self):
         raise ConnectionFailure(
             'Timed out waiting for socket from pool with max_size %r and'
